import posixpath

from fred.utils.imout.interface import ImageOutputInterface


class OutputMinio(ImageOutputInterface):
    """MinIO image output handling."""

    @classmethod
    def auto(cls, **kwargs) -> "OutputMinio":
        from fred.dao.service.catalog import ServiceCatalog
        from fred.dao.comp.catalog import CompCatalog

        minio = ServiceCatalog.MINIO.auto(**kwargs)
        if "minio_endpoint" in minio.metadata:
            cls.metadata["minio_endpoint"] = minio.metadata["minio_endpoint"]
        cls.client = CompCatalog.KEYVAL.mount(minio)
        return cls(**kwargs)

    def out(self, bucket: str, filename: str, presigned: bool = False, **kwargs) -> str:
        from fred.utils.imops import image_to_b64

        image_string = image_to_b64(self.image)
        self.client(key=filename).set(image_string, b64=True, bucket=bucket, **kwargs)
        if not presigned:
            return posixpath.join(
                self.metadata.get("minio_endpoint", ""),
                bucket,
                filename,
            )
        # TODO: Implement the generation of a pre-signed URL
        # https://github.com/minio/minio-py/blob/master/docs/API.md#get_presigned_url
        raise NotImplementedError("Presigned URL generation is not yet implemented.")



from dataclasses import dataclass
from typing import Callable, Optional

from fred.worker.runner.backend import RunnerBackend
from fred.worker.runner.settings import FRD_RUNNER_BACKEND
from fred.rest.router.endpoint import RouterEndpoint
from fred.settings import logger_manager

from fastapi import APIRouter

logger = logger_manager.get_logger(name=__name__)


class RouterInterfaceMixin:
    """Base class for router interfaces that require a backend service."""
    runner_backend: RunnerBackend


@dataclass(frozen=True, slots=True)
class RouterInterface(RouterInterfaceMixin):
    router: APIRouter

    @classmethod
    def set_backend(
            cls,
            service_name: Optional[str] = None,
            disable_runner_reuse: bool = False,
            **kwargs
        ) -> type["RouterInterface"]:
        runner_backend = kwargs.pop("runner_backend", None)
        if not disable_runner_reuse and runner_backend is None:
            logger.warning(
                "Runner reuse is enabled but no existing backend was provided; "
                "a new backend will be created."
            )
            disable_runner_reuse = True
        if disable_runner_reuse:
            runner_backend = RunnerBackend.auto(
                service_name=service_name or FRD_RUNNER_BACKEND,
                **kwargs,
            )
        if runner_backend is None:
            raise ValueError("Runner backend could not be determined or created.")
        return type(
            "RouterInterfaceWithBackend",
            (cls, ),
            {
                "runner_backend": runner_backend
            }
        ) 

    @classmethod
    def auto(
        cls,
        service_name: Optional[str] = None,
        router: Optional[APIRouter] = None,
        disregard_backend: bool = False,
        disregard_endpoint_setup: bool = False,
        **kwargs,
    ) -> "RouterInterface":
        subcls = cls if disregard_backend else cls.set_backend(
            service_name=service_name,
            runner_backend=getattr(cls, "runner_backend", None),
            **kwargs
        )
        instance = subcls(router=router or APIRouter())
        if disregard_endpoint_setup:
            return instance
        for endpoint in subcls.endpoint_definitions():
            instance.register_endpoint(endpoint)
        return instance

    def register_endpoint(self, endpoint: RouterEndpoint, path: Optional[str] = None):
        route_config = endpoint.route_config(path=path)
        self.router.add_api_route(endpoint=endpoint.function, **route_config)

    @classmethod
    def endpoint_definitions(cls) -> list[RouterEndpoint]:
        return [
            obj
            for attr in dir(cls)
            if isinstance((obj := getattr(cls, attr)), RouterEndpoint)
        ]
    
    @classmethod
    def endpoint_definition_mapping(cls) -> dict[str, RouterEndpoint]:
        return {
            endpoint.get_path(): endpoint
            for endpoint in cls.endpoint_definitions()
        }



from fred.rest.router.interface import RouterInterface


class RouterCatalogInterface:

    def auto(self, **kwargs) -> RouterInterface:
        return self.value.reference.auto(**kwargs)

    def get_kwargs(self) -> dict:
        # This method can be overridden to provide specific kwargs for router initialization
        return {}



from dataclasses import dataclass
from typing import Iterator, Optional

from fred.settings import logger_manager, get_environ_variable
from fred.dao.service.catalog import ServiceCatalog
from fred.dao.comp.interface import ComponentInterface

logger = logger_manager.get_logger(name=__name__)


def _get_minio_elements_from_key(key: str, **kwargs) -> tuple[str, str]:
    import os

    for bucket_key in ("bucket", "minio_bucket", "bucket_name"):
        if bucket_key in kwargs:
            bucket_name_candidate = kwargs.pop(bucket_key)
            break
    else:
        bucket_name_candidate = get_environ_variable("MINIO_BUCKET") or os.path.dirname(key)

    fullpath = (key if key.startswith(bucket_name_candidate + os.sep) else os.path.join(bucket_name_candidate, key)).strip(os.sep)

    # We only want the first component as bucket name, the rest is the object name
    bucket_name, *obj_components = os.path.normpath(fullpath).split(os.sep)
    object_name = os.path.join(*obj_components)
    if not bucket_name:
        raise ValueError(
            "Bucket name must be specified either in kwargs, environment variable MINIO_BUCKET, "
            "or implicitly as part of the key."
        )
    if not object_name:
        raise ValueError("Object name cannot be empty.")
    return bucket_name, object_name


@dataclass(frozen=True, slots=True)
class FredKeyVal(ComponentInterface):
    """A simple key-value store implementation using a backend service.
    This class provides methods to interact with a key-value store, such as setting,
    getting, and deleting key-value pairs. The actual implementation of these methods
    depends on the underlying service being used (e.g., Redis).
    """
    key: str
    
    @classmethod
    def keys(cls, pattern: Optional[str] = None, **kwargs) -> Iterator[str]:
        """Returns a list of keys matching the given pattern.
        The implementation of this method depends on the underlying service.
        For example, if the service is Redis, it uses the KEYS command to get the
        list of keys matching the pattern.
        Args:
            pattern (str): The pattern to match keys against. Defaults to "*".
        Returns:
            list[str]: A list of keys matching the pattern.
        Raises:
            NotImplementedError: If the method is not implemented for the current service.
        """
        match cls._cat:
            case ServiceCatalog.REDIS:
                return (
                    key if isinstance(key, str) else key.decode("utf-8")
                    for key in cls._srv.client.scan_iter(match=pattern, **kwargs)
                )
            case ServiceCatalog.STDLIB:
                import fnmatch
                pattern = pattern or "*"
                return (
                    key
                    for key in cls._srv.client._memstore_keyval.keys()
                    if fnmatch.fnmatch(key, pat=pattern)
                )
            case ServiceCatalog.MINIO:
                import fnmatch
                pattern = pattern or "*"
                bucket_name = (
                    kwargs.get("bucket", None)
                    or kwargs.get("minio_bucket", None)
                    or get_environ_variable("MINIO_BUCKET")
                )
                if not bucket_name:
                    raise ValueError("Missing bucket info to list keys in MinIO service.")
                obj_list_extras = {
                    "prefix": kwargs.pop("prefix", ""),
                    "shallow": kwargs.pop("shallow", False),
                }
                return (
                    key
                    for key in cls._srv.objects(bucket_name, **obj_list_extras)
                    if fnmatch.fnmatch(key, pat=pattern)
                )
            case _:
                raise NotImplementedError(f"Keys method not implemented for service {cls._nme}")
        if kwargs:
            logger.warning(f"Additional kwargs ignored: {kwargs}")

    def set(self, value: str, key: Optional[str] = None, **kwargs) -> None:
        """Sets a key-value pair in the store.
        The implementation of this method depends on the underlying service.
        For example, if the service is Redis, it uses the SET command to store the
        key-value pair.
        Args:
            key (str): The key to set.
            value (str): The value to associate with the key.
            **kwargs: Additional keyword arguments for setting the key-value pair,
                      such as expiration time.
        Raises:
            NotImplementedError: If the method is not implemented for the current service.
        """
        key = key or self.key
        match self._cat:
            case ServiceCatalog.REDIS:
                self._srv.client.set(key, value)
                expire = kwargs.get("expire")
                if expire and isinstance(expire, int):
                    self._srv.client.expire(key, expire)
            case ServiceCatalog.STDLIB:
                self._srv.client._memstore_keyval[key] = value
                # TODO: Implement expiration logic
                if "expire" in kwargs:
                    logger.warning("Expiration not implemented for STDLIB service.")
            case ServiceCatalog.MINIO:
                # MinIO is not a key-value store, but we can simulate it:
                # the key will be the object name, and the value will be the object content.
                import io

                bucket_name, object_name = _get_minio_elements_from_key(key, **kwargs)
                # Ensure the bucket exists or create otherwise
                if not self._srv.bucket_exists(bucket_name):
                    logger.warning(f"Creating bucket since doesn't exists: {bucket_name}")
                    self._srv.client.make_bucket(bucket_name)
                if "expire" in kwargs:
                    # TODO: Implement expiration logic
                    logger.warning("Expiration not implemented for MINIO service.")
                # Prepare the value as a byte stream
                value_bytes = value.encode("utf-8")
                if kwargs.get("b64", False):
                    import base64
                    value_bytes = base64.b64decode(value)
                value_stream = io.BytesIO(value_bytes)
                value_stream.seek(0)  # Ensure the stream is at the beginning
                # Put the object into the bucket
                self._srv.client.put_object(
                    bucket_name=bucket_name,
                    object_name=object_name,
                    data=value_stream,
                    length=len(value_bytes),
                )
            case _:
                raise NotImplementedError(f"Set method not implemented for service {self._nme}")
        if kwargs:
            logger.warning(f"Additional kwargs ignored: {kwargs}")

    def get(self, key: Optional[str] = None, fail: bool = False, **kwargs) -> Optional[str]:
        """Gets the value associated with a key from the store.
        The implementation of this method depends on the underlying service.
        For example, if the service is Redis, it uses the GET command to retrieve the
        value associated with the key.
        Args:
            key (str): The key to retrieve.
            fail (bool): If True, raises a KeyError if the key is not found. Defaults to False.
        Returns:
            Optional[str]: The value associated with the key, or None if the key is not found
                           and fail is False.
        Raises:
            KeyError: If the key is not found and fail is True.
            NotImplementedError: If the method is not implemented for the current service.
        """
        key = key or self.key
        result = None
        match self._cat:
            case ServiceCatalog.REDIS:
                result = self._srv.client.get(key)
            case ServiceCatalog.STDLIB:
                result = self._srv.client._memstore_keyval.get(key)
            case ServiceCatalog.MINIO:
                bucket_name, object_name = _get_minio_elements_from_key(key)
                # Verify if the bucket exists
                if not self._srv.bucket_exists(bucket_name):
                    logger.warning(f"Bucket {bucket_name} does not exist.")
                    if fail:
                        raise KeyError(f"Bucket {bucket_name} does not exist.")
                    return None
                # Verify if the object exists
                if not self._srv.object_exists(bucket_name, object_name):
                    logger.warning(f"Object {object_name} in bucket {bucket_name} does not exist.")
                    if fail:
                        raise KeyError(f"Object {object_name} not found in bucket {bucket_name}.")
                    return None
                try:
                    if kwargs.pop("presigned_url", False):
                        result = self._srv.object_presigned_url(bucket_name, object_name, **kwargs)
                    else:
                        response = self._srv.client.get_object(bucket_name, object_name)
                        result_bytes = response.read()
                        try:
                            # This should work for most cases where original text data was stored (e.g., JSON, YAML, CSVs, etc.)
                            result = result_bytes.decode("utf-8")
                        except UnicodeDecodeError:
                            import base64
                            # This should work for binary data (e.g., images, PDFs, etc.)
                            result = base64.b64encode(result_bytes).decode("ascii")
                        finally:
                            response.close()
                            response.release_conn()
                except Exception as e:
                    logger.error(f"Error retrieving object {object_name} from bucket {bucket_name}: {e}")
                    result = None
                    if fail:
                        raise KeyError(f"Object {object_name} not found in bucket {bucket_name}.")
            case _:
                raise NotImplementedError(f"Get method not implemented for service {self._nme}")
        if fail and result is None:
            raise KeyError(f"Key {key} not found.")
        if kwargs:
            logger.warning(f"Additional kwargs ignored: {kwargs}")
        return result

    def delete(self, key: Optional[str] = None) -> None:
        """Deletes a key-value pair from the store.
        The implementation of this method depends on the underlying service.
        For example, if the service is Redis, it uses the DEL command to remove the
        key-value pair.
        Args:
            key (str): The key to delete.
        Raises:
            NotImplementedError: If the method is not implemented for the current service.
        """
        key = key or self.key
        match self._cat:
            case ServiceCatalog.REDIS:
                self._srv.client.delete(key)
            case ServiceCatalog.STDLIB:
                self._srv.client._memstore_keyval.pop(key, None)
            case ServiceCatalog.MINIO:
                bucket_name, object_name = _get_minio_elements_from_key(key)
                # Verify if the bucket exists
                if not self._srv.bucket_exists(bucket_name):
                    logger.warning(f"Bucket {bucket_name} does not exist.")
                    return
                # Verify if the object exists
                if not self._srv.object_exists(bucket_name, object_name):
                    logger.warning(f"Object {object_name} in bucket {bucket_name} does not exist.")
                    return
                try:
                    self._srv.client.remove_object(bucket_name, object_name)
                except Exception as e:
                    logger.error(f"Error deleting object {object_name} from bucket {bucket_name}: {e}")
            case _:
                raise NotImplementedError(f"Delete method not implemented for service {self._nme}")



from typing import Callable, Optional

from fred.edag.comp.catalog import CompCatalog


class NodeDecorator:
    """Decorator to create Node instances from functions.
    Can be used with or without parameters.

    Example usage:

    @NodeDecorator
    def my_function(...):
        ...
    or
    @NodeDecorator(name="example", param1=value1, param2=value2)
        def my_function(...):
    """

    def __new__(cls, func: Optional[Callable] = None, **kwargs):
        if not func:
            # Create an instance of NodeDecorator to hold kwargs until __call__ is invoked
            instance = super(NodeDecorator, cls).__new__(cls)
            instance.kwargs = kwargs
            return instance
        # Return a Node instance directly if func is provided
        return CompCatalog.NODE.ref.auto(
            function=func,
            **kwargs,
        )

    def __init__(self, func: Optional[Callable] = None, **kwargs):
        self.func = func
        self.kwargs = kwargs

    def __call__(self, func: Callable, **kwargs) -> CompCatalog.NODE.ref:
        self.func = func
        self.kwargs = {**self.kwargs, **kwargs}
        return self.get_node()

    def get_node(self, **kwargs) -> CompCatalog.NODE.ref:
        return CompCatalog.NODE.ref.auto(
            function=self.func,
            **{
                **self.kwargs,
                **kwargs,
            },
        )



import enum

from fred.edag.comp.interface import ComponentInterface
from fred.edag.comp._group import Group
from fred.edag.comp._node import Node
from fred.edag.plan import Plan


class PlanCompositionMixin:

    def __rshift__(self, other: ComponentInterface | Plan) -> Plan:
        left = Plan.as_plan(self)
        right = Plan.as_plan(other)
        return left >> right


class CompCatalog(enum.Enum):
    NODE = type("Node", (Node, PlanCompositionMixin), {})
    GROUP = type("Group", (Group, PlanCompositionMixin), {})

    @property
    def ref(self) -> type[ComponentInterface]:
        return self.value

    def __call__(self, *args, **kwargs) -> ComponentInterface:
        return self.value(*args, **kwargs)



from typing import Optional

from fred.version import version
from fred.settings import logger_manager
from fred.cli.interface import AbstractCLI


logger = logger_manager.get_logger(name=__name__)


class CLIExtensionGroups:
    """CLI Extensions providing access to various integrations by following a lazy loading pattern."""

    @property
    def databricks(self):
        from fred.integrations.databricks.cli_ext import DatabricksExt
        return DatabricksExt()
    
    @property
    def runpod(self):
        from fred.integrations.runpod.cli_ext import RunPodExt
        return RunPodExt()
    
    @property
    def runner_server(self):
        from fred.worker.runner.rest.cli_ext import RunnerServerExt
        return RunnerServerExt()


class CLI(AbstractCLI, CLIExtensionGroups):

    def version(self) -> str:
        return version.value

    def serve(
            self,
            classname: Optional[str] = None,
            classpath: Optional[str] = None,
            include_routers: Optional[list[str]] = None,
            exclude_routers: Optional[list[str]] = None,
            fastapi_configs: Optional[dict] = None,
            server_configs: Optional[dict] = None,
    ):
        from fred.rest.server import FredServer

        include_routers = include_routers or []
        exclude_routers = exclude_routers or []
        fastapi_configs = fastapi_configs or {}
        server_configs = server_configs or {}

        logger.info("Starting the Fred-REST Server...")
        server = FredServer.auto(
            include_routers=include_routers,
            exclude_routers=exclude_routers,
            router_classname=classname,
            router_classpath=classpath,
            **fastapi_configs,
        )
        server.start(**server_configs)



"""
[Fred-REST] Fred's Simple Extensible REST-API Server
========================================================
This module provides a simple and extensible REST-API server implementation using FastAPI.
It includes configurations for authentication, router management, and server startup.
"""
from fred.maturity import Maturity, MaturityLevel


module_maturity = Maturity(
    level=MaturityLevel.ALPHA,
    reference=__name__,
    message=(
        "Fred-REST implementation is in early development "
        "and therefore currently with incomplete and unstable features."
    )
)



from fred.utils.imout.interface import ImageOutputInterface


class OutputFilesystem(ImageOutputInterface):
    """Filesystem output handler for images."""

    def out(self, path: str) -> str:
        from fred.utils.imops import save_image_to_path

        save_image_to_path(image=self.image, path=path)
        return path


import enum


class MinioPolicyBuilderActionBucket(enum.StrEnum):
    ALL = "s3:*"
    LIST = "s3:ListBucket"
    LOCATION = "s3:GetBucketLocation"
    DELETE = "s3:DeleteBucket"
    CREATE = "s3:CreateBucket"


class MinioPolicyBuilderActionObject(enum.StrEnum):
    GET = "s3:GetObject"
    PUT = "s3:PutObject"
    DELETE = "s3:DeleteObject"
    ALL = "s3:*"



import uuid
from dataclasses import dataclass
from typing import Optional

from fred.settings import logger_manager
from fred.dao.service.catalog import ServiceCatalog
from fred.dao.comp.interface import ComponentInterface

logger = logger_manager.get_logger(name=__name__)


class FredSubscriptionMixin:
    # TODO: Improve typing... and better underlying subscriptions management!
    # This only works within the same python process.
    subs: dict = {}


@dataclass(frozen=True, slots=True)
class FredPubSub(ComponentInterface, FredSubscriptionMixin):
    """A simple publish-subscribe (pub-sub) implementation using a backend service.
    This class provides methods to interact with pub-sub channels, such as publishing
    messages, subscribing to channels, and managing subscriptions. The actual implementation
    of these methods depends on the underlying service being used (e.g., Redis).
    Attributes:
        name: str: The name of the pub-sub channel.
    """
    name: str

    def publish(self, item: str) -> int:
        match self._cat:
            case ServiceCatalog.REDIS:
                return self._srv.client.publish(self.name, item)
            case ServiceCatalog.STDLIB:
                raise NotImplementedError("Publish method not implemented for STDLIB service")
            case _:
                raise NotImplementedError(f"Publish method not implemented for service {self._nme}")

    def subscribe(self, subscription_id: Optional[str] = None):
        """Subscribe to the pub/sub channel and yield messages as they arrive.

        This method creates (or reuses) a subscription to the channel specified by `self.name`.
        It returns a generator that yields messages received on the channel.
        The implementation depends on the underlying service (e.g., Redis).

        Args:
            subscription_id (Optional[str]): An optional identifier for the subscription. If not provided, a new UUID is generated.

        Yields:
            Messages received from the channel.
        Raises:
            NotImplementedError: If the method is not implemented for the current service.
        """
        subscription_id = subscription_id or (
            logger.info(f"Creating new subscriber for channel: {self.name}")
            or str(uuid.uuid4())
        )
        logger.info(f"Using subscription ID: {subscription_id}")
        match self._cat:
            case ServiceCatalog.REDIS:
                subscriber = self.subs[subscription_id] = self.subs.get(subscription_id, None) or self._srv.client.pubsub()
                subscriber.subscribe(self.name)
                yield from subscriber.listen()
            case ServiceCatalog.STDLIB:
                raise NotImplementedError("Subscribe method not implemented for STDLIB service")
            case _:
                raise NotImplementedError(f"Subscribe method not implemented for service {self._nme}")

    @classmethod
    def unsubscribe(cls, subscription_id: str, close: bool = False) -> None:
        # TODO: Implement the 'unsubscribe' and optially close the subscriber.
        logger.error("The unsubscribe method not implemented yet.")
        raise NotImplementedError("Unsubscribe method not implemented yet.")

    @classmethod
    def subscribers(cls) -> list[str]:
        # TODO: Implement the 'subscribers' method to list active subscriptions.
        logger.error("The subscribers method not implemented yet.")
        raise NotImplementedError("Subscribers method not implemented yet.")



from dataclasses import dataclass

from fred.edag.comp.interface import ComponentInterface
from fred.edag.comp._node import Node


@dataclass(frozen=True, slots=True, unsafe_hash=True)
class Group(ComponentInterface):
    nodes: list[Node]  # TODO: let's make this a frozenset instead of a list to ensure immutability



import enum
from typing import Optional

from fred.dao.service._minio.policy.loader import MinioPolicyLoader


class MinioPolicyCatalog(enum.Enum):
    BUCKET_PUBLIC_RO = MinioPolicyLoader(
        title="[Bucket Policy] Public Read-Only",
        filename="public_ro.json",
        requires=[
            "bucket_name",
        ],
    )
    BUCKET_PUBLIC_RW = MinioPolicyLoader(
        title="[Bucket Policy] Public Read-Write",
        filename="public_rw.json",
        requires=[
            "bucket_name",
        ],
    )

    def load(self, path: Optional[str] = None, **params) -> dict:
        return self.value.load(path=path, **params)
    
    def content(self, path: Optional[str] = None, **params) -> str:
        import json
        return json.dumps(self.load(path=path, **params))



from dataclasses import dataclass
from typing import Optional

from fred.rest.router.catalog.interface import RouterCatalogInterface
from fred.rest.settings import (
    FRD_RESTAPI_ROUTERCATALOG_CLASSPATH,
    FRD_RESTAPI_ROUTERCATALOG_CLASSNAME,
)

@dataclass(frozen=True, slots=True)
class ServerRouterCatalogConfig:
    classname: str
    classpath: str

    @classmethod
    def auto(cls, classname: Optional[str] = None, classpath: Optional[str] = None) -> "ServerRouterCatalogConfig":
        return cls(
            classname=classname or FRD_RESTAPI_ROUTERCATALOG_CLASSNAME,
            classpath=classpath or FRD_RESTAPI_ROUTERCATALOG_CLASSPATH,
        )

    @property
    def catalog(self) -> type[RouterCatalogInterface]:
        import importlib

        module = importlib.import_module(self.classpath)
        catalog_class = getattr(module, self.classname, None)
        if catalog_class is None:
            raise ImportError(f"Could not find class '{self.classname}' in module '{self.classpath}'")
        if not issubclass(catalog_class, RouterCatalogInterface):
            raise TypeError(f"Class '{self.classname}' is not a subclass of RouterCatalogInterface")
        return catalog_class



from dataclasses import dataclass

from fred.rest.router.interface import RouterInterface, RouterInterfaceMixin


@dataclass(frozen=True, slots=True)
class RouterConfig:
    prefix: str
    other: dict

    @classmethod
    def auto(cls, prefix: str = "", **kwargs) -> "RouterConfig":
        return cls(
            prefix=prefix,
            other=kwargs,
        )

    def get_applied_configs(self, router_class: type[RouterInterface]) -> "AppliedRouterConfig":
        return AppliedRouterConfig(
            reference=router_class,
            config=self,
        )

    def __call__(self, apply: type[RouterInterfaceMixin]) -> "AppliedRouterConfig":
        router_classname = f"Router{apply.__name__}"
        return self.get_applied_configs(
            router_class=type(router_classname, (RouterInterface, apply, ), {})
        )
    
    def get_configs(self) -> dict:
        return {
            "prefix": self.prefix,
            **self.other,
        }


@dataclass(frozen=True, slots=True)
class AppliedRouterConfig:
    reference: type[RouterInterface]
    config: RouterConfig



from fastapi import Security, HTTPException, status
from fastapi.security import APIKeyHeader

from fred.rest.settings import FRD_RESTAPI_TOKEN


async def verify_key(
        api_key_header: str = Security(
            APIKeyHeader(
                name="X-API-Key",
                auto_error=False
            )
        ),
):
    """
    Verify the provided API key against the expected token.
    Raises an HTTPException if the key is invalid or missing.
    """
    if api_key_header != FRD_RESTAPI_TOKEN:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid or missing API Key",
        )



from typing import Optional

from fred.future import Future
from fred.settings import logger_manager
from fred.utils.dateops import datetime_utcnow
from fred.rest.router.interface import RouterInterfaceMixin
from fred.rest.router.endpoint import RouterEndpointAnnotation

logger = logger_manager.get_logger(name=__name__)


class RunnerRouterMixin(RouterInterfaceMixin):

    @RouterEndpointAnnotation.set(
        path="/handler_exists",
        methods=["GET"],
        tags=["Runner"],
        summary="Check if a handler class exists and is a RunnerHandler.",
        response_description="Details about the handler class.",
    )
    def handler_exists(self, classname: str, classpath: str, **kwargs) -> dict:
        from fred.worker.runner.handler import RunnerHandler
        from fred.worker.interface import HandlerInterface

        result_payload = {
            "handler_classname": classname,
            "handler_classpath": classpath,
            "exists": False,
            "is_runner_handler": False,
            "metadata": {}
        }

        try:
            handler = HandlerInterface.find_handler(
                import_pattern=classpath,
                handler_classname=classname,
            )
            result_payload["is_runner_handler"] = isinstance(handler, RunnerHandler)
            result_payload["exists"] = True
            return result_payload
        except Exception as e:
            result_payload["metadata"]["error"] = str(e)
            return result_payload


    @RouterEndpointAnnotation.set(
        path="/qlen/{queue_slug}",
        methods=["GET"],
        tags=["Runner"],
        summary="Get the length of the request and response queues for a given queue slug.",
        response_description="The lengths of the request and response queues.",
    )
    def qlen(self, queue_slug: str, **kwargs) -> dict:
        snapshot_at = datetime_utcnow().isoformat()
        req_queue = self.runner_backend.queue(f"req:{queue_slug}")
        res_queue = self.runner_backend.queue(f"res:{queue_slug}")
        backend_name = self.runner_backend._cat.name
        return {
            "snapshot_at": snapshot_at,
            "queue_slug": queue_slug,
            "queue_backend": backend_name,
            "req": req_queue.size(),
            "res": res_queue.size(),
        }

    @RouterEndpointAnnotation.set(
        path="/start",
        methods=["POST"],
        tags=["Runner"],
        summary="Start a runner using the specified plugin.",
        response_description="The ID of the started runner.",
    )
    def runner_start(self, **kwargs) -> dict:
        from fred.worker.runner.model.catalog import RunnerModelCatalog
        from fred.worker.runner.plugins.catalog import PluginCatalog
        # Determine which plugin to use; default to LOCAL if not specified
        plugin_name: str = kwargs.pop("plugin", "LOCAL")
        wait_for_exec: bool = kwargs.pop("wait_for_exec", False)
        # Create the RunnerSpec from the provided payload
        # TODO: Instead on depending on parsing a dict... Can we implement a base-model to facilitate fast-api validation?
        runner_spec = RunnerModelCatalog.RUNNER_SPEC.value.from_payload(payload=kwargs)
        # Instantiate the plugin and execute the runner
        plugin = PluginCatalog[plugin_name.upper()]()
        output = plugin.execute(runner_spec, wait_for_exec=wait_for_exec, **kwargs)
        return {
            "runner_id": output.runner_id,
            "future_id": output.future_exec.future_id,
            "queue_slug": runner_spec.queue_slug,
        }

    @RouterEndpointAnnotation.set(
        path="/execute",
        methods=["POST"],
        tags=["Runner"],
        summary="Execute a task by dispatching a request to the specified queue.",
        response_description="Details about the dispatched request.",
    )
    def runner_execute(self, **kwargs) -> dict:
        from fred.worker.runner.model.catalog import RunnerModelCatalog

        request_id = kwargs.pop("request_id", None)
        queue_slug = kwargs.pop("queue_slug", None) or (
            logger.error("No 'queue_slug' value provided; defaulting to 'demo'.")
            or "demo"
        )

        item = RunnerModelCatalog.ITEM.value.uuid(payload=kwargs, uuid_hash=False)
        request = item.as_request(use_hash=False, request_id=request_id)
        request.dispatch(
            request_queue=self.runner_backend.queue(f"req:{queue_slug}")
        )
        # Starting the runner to process the request if requested; this should always be BEFORE placing the request in the queue
        # to avoid race conditions where a blocking runner is spawned before the request is enqueued.
        # TODO: Let's optimize this by identifying if there's already an on-going runner for the same spec.
        # If so, we can potentially reuse it instead of starting a new one each time.
        # TODO: Let's remove the '_og' reference and find a cleaner way to handle this...
        runner_start_output = self.runner_start._og(self, **start_configs) \
            if (start_configs := kwargs.pop("start_configs", {})) else {}
        return {
            "item_id": item.item_id,
            "request_id": request.request_id,
            "queue_slug": queue_slug,
            "dispatched_at": datetime_utcnow().isoformat(),
            "runner_start_output": runner_start_output,
        }

    @RouterEndpointAnnotation.set(
        path="/output/{request_id}",
        methods=["GET"],
        tags=["Runner"],
        summary="Fetch the output of a previously dispatched request.",
        response_description="The output of the request.",
    )
    def runner_output(self, request_id: str, nonblocking: bool = False, timeout: Optional[float] = None, **kwargs) -> dict:

        output_requested_at = datetime_utcnow().isoformat()
        # Subscribe to the future result using the request_id
        future = Future.subscribe(future_id=request_id)
        if nonblocking:
            # TODO: Implement non-blocking fetch logic... let's ensure it's worth the effort.
            raise NotImplementedError("Non-blocking fetch not implemented yet...")
        return {
            "request_id": request_id,
            "output_requested_at": output_requested_at,
            "output_delivered_at": datetime_utcnow().isoformat(),
            "output": future.wait_and_resolve(timeout=timeout),
        }



from dataclasses import dataclass, field
from typing import FrozenSet, Tuple, Union

from fred.edag.comp._node import Node
from fred.edag.comp._group import Group
from fred.edag.comp.interface import ComponentInterface


@dataclass(frozen=True, slots=True)
class Plan:
    nodes: list[Node] = field(default_factory=list)
    edges: list[Tuple[Node, Node]] = field(default_factory=list)  # directed edges: (from/src, to/dst)
    heads: list[Node] = field(default_factory=list)  # current entry set of chaining
    tails: list[Node] = field(default_factory=list)  # current exit set of chaining

    @staticmethod
    def empty() -> 'Plan':
        return Plan()

    @classmethod
    def as_plan(cls, other: Union[ComponentInterface, 'Plan']) -> 'Plan':
        match other:
            case Plan() as plan:
                return plan
            case Node():
                return Plan(
                    nodes=[other],
                    edges=[],
                    heads=[other],
                    tails=[other],
                )
            case Group():
                return Plan(
                    nodes=[*other.nodes],
                    edges=[],
                    heads=[*other.nodes],
                    tails=[*other.nodes],
                )
            case _:
                raise TypeError("as_plan expects Node, Group, or Plan")

    def __rshift__(self, other: Union[ComponentInterface, 'Plan']) -> "Plan":
        other_plan = self.as_plan(other)
        # combine edges, nodes, and link tails -> heads (i.e., connect current tails to new heads)
        new_edges = [
            *self.edges,
            *other_plan.edges,
            *(
                (src, dst)
                for src in self.tails
                for dst in other_plan.heads
            ),
        ]
        new_nodes = [*self.nodes, *other_plan.nodes]
        return Plan(
            nodes=new_nodes,
            edges=new_edges,
            heads=self.heads,
            tails=other_plan.tails,
        )

    def as_predmap(self) -> dict[Node, set[Node]]:
        predmap = {n: set() for n in self.nodes}  # Ensure all nodes are keys
        for src, dst in self.edges:
            predmap[dst].add(src)
        return predmap



from dataclasses import dataclass

from fred.rest.router.interface import RouterInterfaceMixin
from fred.rest.router.endpoint import RouterEndpointAnnotation

@dataclass(frozen=True, slots=True)
class RouterBaseMixin(RouterInterfaceMixin):
    
    @RouterEndpointAnnotation.set(
        path="/",
        methods=["GET"],
        summary="Base endpoint",
        description="A simple base endpoint to check service availability."
    )
    def base(self, include_telemetry: bool = False, **kwargs) -> dict:
        if include_telemetry:
            from fred.utils.runtime import RuntimeProfilingSnapshot
            kwargs["telemetry"] = RuntimeProfilingSnapshot.auto().to_dict()
        return {
            "ok": True,
            **kwargs
        }



from dataclasses import dataclass
from typing import Optional

from fred.rest.router.interface import RouterInterfaceMixin
from fred.rest.router.endpoint import RouterEndpointAnnotation


class RouterExampleMixin(RouterInterfaceMixin):

    @RouterEndpointAnnotation.set(
        path="/ping",
        methods=["GET"],
        summary="Ping endpoint",
        description="A simple ping endpoint to check service availability.",
    )
    def ping(self, pong: Optional[str] = None, **kwargs) -> dict:
        from fred.utils.dateops import datetime_utcnow

        return {
            "ping_time": datetime_utcnow().isoformat(),
            "ping_response": pong or "pong",
        }

    @RouterEndpointAnnotation.set(
        path="/passthrough",
        methods=["GET"],
        summary="Passthrough GET endpoint",
        description="An endpoint that returns all received parameters.",
    )
    def passthrough(self, **kwargs) -> dict:
        return kwargs



import enum

from fred.rest.router.config import RouterConfig
from fred.rest.router.catalog.interface import RouterCatalogInterface
from fred.worker.runner.rest.router._runner import RunnerRouterMixin
from fred.worker.runner.rest.router._base import RouterBaseMixin


class RouterCatalog(RouterCatalogInterface, enum.Enum):
    BASE = RouterConfig.auto(prefix="")(apply=RouterBaseMixin)
    RUNNER = RouterConfig.auto(prefix="/runner", tags=["Runner"])(apply=RunnerRouterMixin)



import uuid
from graphlib import TopologicalSorter
from dataclasses import dataclass, field
from typing import Any, Optional

from fred.future.impl import Future
from fred.edag.comp.catalog import CompCatalog
from fred.edag.plan import Plan


@dataclass(frozen=True, slots=True)
class Executor:
    predmap: dict[CompCatalog.NODE.ref, set[CompCatalog.NODE.ref]]
    results: dict[str, dict[str, Any]] = field(default_factory=dict)

    @classmethod
    def from_plan(cls, plan: Plan, **kwargs) -> "Executor":
        return cls(predmap=plan.as_predmap(**kwargs))
    
    def get_tsort(self) -> TopologicalSorter:
        return TopologicalSorter(self.predmap)
        
    def loop(
            self,
            run_id: str,
            tsort: TopologicalSorter,
            prev_layer: list[list[str]],
            start_with: Optional[dict] = None,
            unrestricted: bool = False,
            non_destructive_node_explosion: bool = False,
    ) -> list[list[str]]:
        start_with = start_with or {}
        if not (nodes := tsort.get_ready()):
            return prev_layer
        # You can only get access to results of previous layers unless unrestricted is requested.
        prev_layer_results = self.results[run_id] if unrestricted else {
            key: val
            for key, val in self.results[run_id].items()
            if key in prev_layer[-1]
        }
        curr_layer = []
        for node in nodes:
            parents = [
                parent.name
                for parent in self.predmap.get(node, [])
            ]
            accessible_results = prev_layer_results if unrestricted else {
                key: val
                for key, val in prev_layer_results.items()
                if key in parents
            }
            kwargs = {
                **start_with,
                **{
                    arg: val
                    for node_key, node_out in accessible_results.items()
                    for arg, val in node_out.items()
                }
            }
            # Execute node function
            match node.execute(**kwargs):
                case Future() as future:
                    # Can't we just build the whole graph in the future and 'wait_and_resolve' only at the end?
                    # Or at least per layer/generation?
                    output = {node.key: future.wait_and_resolve()}
                case present:
                    output = {node.key: present}
            # We can only explode if  requested and the output result is a dict
            if node._explode and isinstance(output.get(node.key), dict):
                output = {
                    # Keep original output if non-destrictive-explode is requested;
                    # The original key can be overwritten if key collides during explosion.
                    **(output if non_destructive_node_explosion else {}),
                    # Explode keys into the output dict
                    **output[node.key],
                }
            # Store output in results
            self.results[run_id][node.name] = {
                #**self.results[run_id].get(node.name, {}),
                **output,
            }
            # Mark node as done
            tsort.done(node)
            curr_layer.append(node.name)
        prev_layer.append(curr_layer)
        return self.loop(
            run_id=run_id,
            tsort=tsort,
            prev_layer=prev_layer,
            unrestricted=unrestricted,
            start_with={},  # Only availabe during the first layer call
            non_destructive_node_explosion=non_destructive_node_explosion,
        )

    def execute(
            self,
            keep: bool = False,
            unrestricted: bool = False,
            start_with: Optional[dict] = None,
            non_destructive_node_explosion: bool = False,
        ) -> dict:
        from fred.utils.dateops import datetime_utcnow

        run_id = str(uuid.uuid4())
        run_start = datetime_utcnow()
        # Initialize in-memory result storage for this run
        # TODO: Swap the result-store to our fred-keyval implementation
        self.results[run_id] = {}
        # Prepare TopologicalSorter
        tsort = self.get_tsort()
        tsort.prepare()
        # Execute nodes in topological order
        layers = self.loop(
            run_id=run_id,
            tsort=tsort,
            prev_layer=[[]],
            unrestricted=unrestricted,
            start_with=start_with or {},
            non_destructive_node_explosion=non_destructive_node_explosion,
        )
        return {
            "run_id": run_id,
            "run_start": run_start,
            "run_end": datetime_utcnow(),
            "results": self.results[run_id] if keep else self.results.pop(run_id),
            "layers": layers,
        }



from dataclasses import dataclass

from fred.settings import logger_manager
from fred.dao.comp.catalog import FredKeyVal, FredQueue, CompCatalog
from fred.dao.service.interface import ServiceInterface
from fred.dao.service.catalog import ServiceCatalog

logger = logger_manager.get_logger(name=__name__)


@dataclass(frozen=True, slots=False)
class RunnerBackend:
    keyval: FredKeyVal
    queue: FredQueue
    # NOTE: Catalog and service instance references (for internal use)
    # should be a temporal fix until we refactor the backend system
    # according to a more precise usage-patter (still to be defined/identified).
    _cat: ServiceCatalog  # Contains complementary service info and simple references
    _srv: ServiceInterface  # Allows direct access to a client instance

    @classmethod
    def auto(cls, service_name: str, **kwargs) -> 'RunnerBackend':
        match (srv_catalog := ServiceCatalog[service_name.upper()]):
            case ServiceCatalog.REDIS:
                from fred.dao.service.utils import get_redis_configs_from_payload
                service_kwargs = get_redis_configs_from_payload(kwargs)
            case ServiceCatalog.STDLIB:
                service_kwargs = {}
            case _:
                logger.error(
                    f"Unknown service '{service_name}'... "
                    "will attempt to use provided kwargs as-is."
                )
                service_kwargs = kwargs
        logger.info(f"Initializing RunnerBackend using service '{service_name}'")
        srv_instance = srv_catalog.auto(**service_kwargs)
        return cls(
            keyval=CompCatalog.KEYVAL.value.mount(srv_ref=srv_instance),
            queue=CompCatalog.QUEUE.value.mount(srv_ref=srv_instance),
            _cat=srv_catalog,
            _srv=srv_instance,
        )



from fred.settings import get_environ_variable


def get_redis_configs_from_payload(
        payload: dict,
        keep: bool = False,
    ) -> dict:
    """Extract Redis configuration from the given payload dictionary.
    This function looks for common Redis configuration keys in the payload
    dictionary. If a key is not found, it falls back to environment variables.
    Args:
        payload (dict): The dictionary from which to extract Redis configuration.
        keep (bool): If True, the original keys are retained in the payload. If False, they are removed.
    Returns:
        dict: A dictionary containing Redis configuration parameters.
    """
    host = port = password = db = None
    for host_key in ["host", "redis_host"]:
        if (host := payload.get(host_key) if keep else payload.pop(host_key, None)):
            break
    for port_key in ["port", "redis_port"]:
        if (port := payload.get(port_key) if keep else payload.pop(port_key, None)):
            break
    for password_key in ["password", "redis_password"]:
        if (password := payload.get(password_key) if keep else payload.pop(password_key, None)):
            break
    for db_key in ["db", "redis_db"]:
        if (db := payload.get(db_key) if keep else payload.pop(db_key, None)):
            break
    return {
        "host": host or get_environ_variable(name="REDIS_HOST", default=None) or "localhost",
        "port": int(port or get_environ_variable(name="REDIS_PORT", default=None) or 6379),
        "password": password or get_environ_variable(name="REDIS_PASSWORD", default=None),
        "db": int(db or get_environ_variable(name="REDIS_DB", default=None) or 0),
        "decode_responses": True,
        **(payload.get("redis_configs", {}) if keep else payload.pop("redis_configs", {})),
    }


def get_minio_from_payload(
        payload: dict,
        keep: bool = False,
):
    """Extract MinIO configuration from the given payload dictionary.
    This function looks for common MinIO configuration keys in the payload
    dictionary. If a key is not found, it falls back to environment variables.
    Args:
        payload (dict): The dictionary from which to extract MinIO configuration.
        keep (bool): If True, the original keys are retained in the payload. If False, they are removed.
    Returns:
        dict: A dictionary containing MinIO configuration parameters.
    """
    endpoint = access_key = secret_key = region = secure = None
    for endpoint_key in ["endpoint", "minio_endpoint"]:
        if (endpoint := payload.get(endpoint_key) if keep else payload.pop(endpoint_key, None)):
            break
    for access_key_key in ["access_key", "minio_access_key"]:
        if (access_key := payload.get(access_key_key) if keep else payload.pop(access_key_key, None)):
            break
    for secret_key_key in ["secret_key", "minio_secret_key"]:
        if (secret_key := payload.get(secret_key_key) if keep else payload.pop(secret_key_key, None)):
            break
    for secure_key in ["secure", "minio_secure"]:
        if (secure := payload.get(secure_key) if keep else payload.pop(secure_key, None)):
            break
    for region_key in ["region", "minio_region"]:
        if (region := payload.get(region_key) if keep else payload.pop(region_key, None)):
            break
    return {
        "endpoint": endpoint or get_environ_variable(name="MINIO_ENDPOINT", default=None) or "localhost:9000",
        "access_key": access_key or get_environ_variable(name="MINIO_ACCESS_KEY", default=None) or "minioadmin",
        "secret_key": secret_key or get_environ_variable(name="MINIO_SECRET_KEY", default=None) or "minioadmin",
        "secure": bool(secure or get_environ_variable(name="MINIO_SECURE", default=None) or True),  # Default to True
        "region": region or get_environ_variable(name="MINIO_REGION", default=None) or "us-east-1",
        **(payload.get("minio_configs", {}) if keep else payload.pop("minio_configs", {})),
    }



from fred.rest.router.catalog.default.catalog import RouterCatalog
from fred.settings import logger_manager

logger = logger_manager.get_logger(name=__name__)


logger.warning(
    "You are using the default router catalog intended for demonstration purposes only. "
    "For production use, please implement a custom router catalog."
)



import enum

from PIL.Image import Image

from fred.utils.imout.interface import ImageOutputInterface
from fred.utils.imout._filesystem import OutputFilesystem
from fred.utils.imout._string import OutputString
from fred.utils.imout._minio import OutputMinio


class ImageOutputCatalog(enum.Enum):
    B64 = OutputString
    STRING = OutputString
    FILESYSTEM = OutputFilesystem
    MINIO = OutputMinio

    def __call__(self, image: Image, **kwargs) -> ImageOutputInterface:
        if getattr(self.value, "auto", None):
            return self.value.auto(image=image, **kwargs)
        return self.value(image=image, **kwargs)



import enum

from fred.edag.comp.catalog import CompCatalog


class ConnCatalog(enum.Enum):
    PASS = CompCatalog.NODE.ref.auto(
        name="conn:passthrough",
        function=lambda **kwargs: kwargs,
        explode=True,
        inplace=True,
    )
    PASS_PRINT = CompCatalog.NODE.ref.auto(
        name="conn:passthrough_print",
        function=lambda **kwargs: print(kwargs) or kwargs,
        explode=True,
        inplace=True,
    )

    def __call__(self, *args, **kwargs) -> CompCatalog.NODE.ref:
        return self.value.clone(*args, **kwargs)



from fred.settings import get_environ_variable, logger_manager

logger = logger_manager.get_logger(name=__name__)


FRD_RESTAPI_HOST = get_environ_variable(
    "FRD_RESTAPI_HOST",
    default="0.0.0.0"
)
FRD_RESTAPI_PORT = int(get_environ_variable(
    "FRD_RESTAPI_PORT",
    default=8000
))

FRD_RESTAPI_LOGLEVEL = get_environ_variable(
    "FRD_RESTAPI_LOGLEVEL",
    default="info"
).lower()

FRD_RESTAPI_DISABLE_AUTH = get_environ_variable(
    "FRD_RESTAPI_DISABLE_AUTH",
    default="false"
).lower() in ("1", "true", "yes", "on")

FRD_RESTAPI_TOKEN = get_environ_variable(
    "FRD_RESTAPI_TOKEN",
    default=None
)
if not FRD_RESTAPI_TOKEN:
    logger.warning("FRD_RESTAPI_TOKEN not found in environment; using default token 'changeme'.")
    FRD_RESTAPI_TOKEN = "changeme"

FRD_RESTAPI_EXCLUDE_BUILTIN_ROUTERS = get_environ_variable(
    "FRD_RESTAPI_EXCLUDE_BUILTIN_ROUTERS",
    default="false",
).lower() in ("1", "true", "yes", "on")

FRD_RESTAPI_ROUTERCATALOG_CLASSNAME = get_environ_variable(
    "FRD_RESTAPI_ROUTERCATALOG_CLASSNAME",
    default="RouterCatalog",
)

FRD_RESTAPI_ROUTERCATALOG_CLASSPATH = get_environ_variable(
    "FRD_RESTAPI_ROUTERCATALOG_CLASSPATH",
    default="fred.rest.router.catalog.default",
)



import datetime as dt
from typing import Iterator

from minio import Minio

from fred.settings import logger_manager
from fred.dao.service.interface import ServiceInterface
from fred.dao.service.utils import get_minio_from_payload
from fred.dao.service._minio.pool import MinioConnectionPool
from fred.dao.service._minio.policy.catalog import MinioPolicyCatalog

logger = logger_manager.get_logger(name=__name__)


class MinioService(ServiceInterface[Minio]):
    instance: Minio
    metadata: dict = {}

    @classmethod
    def _create_instance(cls, disable_cert: bool = False, **kwargs) -> Minio:
        pool_configs = kwargs.pop("pool_configs", {})
        minio_configs = get_minio_from_payload(kwargs)
        if "http_client" not in minio_configs:
            logger.warning("Creating a new HTTP client for MinIO with connection pooling.")
            minio_configs["http_client"] = MinioConnectionPool.get_or_create_pool(
                disable_cert=disable_cert,
                **pool_configs
            )
        cls.metadata["minio_endpoint"] = minio_configs.get("endpoint")
        return Minio(cert_check=not disable_cert, **minio_configs)

    @classmethod
    def auto(cls, disable_cert: bool = False, **kwargs) -> "MinioService":
        cls.instance = cls._create_instance(disable_cert=disable_cert, **kwargs)
        return cls(**kwargs)

    def buckets(self) -> list[str]:
        """List all buckets in the MinIO instance."""
        return [
            bucket.name
            for bucket in self.client.list_buckets()
        ]

    def objects(self, bucket_name: str, prefix: str = "", shallow: bool = False) -> Iterator[str]:
        """List all objects in a specific bucket in the MinIO instance."""
        return (
            obj.object_name
            for obj in self.client.list_objects(bucket_name, prefix=prefix, recursive=not shallow)
        )

    def bucket_exists(self, bucket_name: str) -> bool:
        """Check if a bucket exists in the MinIO instance."""
        return self.client.bucket_exists(bucket_name)
    
    def object_info(self, bucket_name: str, object_name: str) -> dict:
        """Get metadata of an object in a specific bucket in the MinIO instance."""
        stat = self.client.stat_object(bucket_name, object_name)
        return {
            "bucket_name": bucket_name,
            "object_name": object_name,
            "size": stat.size,
            "last_modified": stat.last_modified,
            "etag": stat.etag,
            "content_type": stat.content_type,
            "metadata": stat.metadata,
        }

    def object_presigned_url(self, bucket_name: str, object_name: str, expiration_hours: int = 6, **kwargs) -> str:
        """Generate a presigned URL for an object in a specific bucket in the MinIO instance."""
        return self.client.presigned_get_object(
            bucket_name=bucket_name,
            object_name=object_name,
            expires=dt.timedelta(hours=expiration_hours),
            **kwargs,
        )

    def object_exists(self, bucket_name: str, object_name: str) -> bool:
        """Check if an object exists in a specific bucket in the MinIO instance."""
        from minio.error import S3Error

        try:
            self.client.stat_object(bucket_name, object_name)
            return True
        except S3Error:
            logger.debug(f"Object {object_name} in bucket {bucket_name} does not exist.")
            return False

    def make_bucket_public(self, bucket_name: str, readonly: bool = False):
        """Make a bucket public with either read-only or read-write access."""
        policy = MinioPolicyCatalog.BUCKET_PUBLIC_RO \
            if readonly else MinioPolicyCatalog.BUCKET_PUBLIC_RW
        self.client.set_bucket_policy(
            bucket_name=bucket_name,
            policy=policy.content(bucket_name=bucket_name)
        )



from urllib3 import PoolManager, Retry
from urllib3.util import Timeout

from fred.settings import get_environ_variable, logger_manager
from fred.dao.service.interface import ServiceConnectionPoolInterface

logger = logger_manager.get_logger(name=__name__)


class MinioConnectionPool(ServiceConnectionPoolInterface[PoolManager]):

    @classmethod
    def _create_pool(cls, disable_cert: bool = False, **kwargs) -> PoolManager:
        """Create a urllib3 PoolManager with the given configurations.

        TODO: Consider using the inverse of 'require_cert' as the default to ensure we do have cert-check automatically.
        For now, we keep it as is to avoid breaking changes.

        Args:
            require_cert (bool): Whether to require SSL certificate verification.
            **kwargs: Additional keyword arguments to pass to the PoolManager constructor.
        Returns:
            PoolManager: A configured PoolManager instance.
        """
        num_pools = kwargs.pop("num_pools", 10)
        maxsize = kwargs.pop("maxsize", 10)
        # Default timeout of 5 minutes
        timeout_seconds = kwargs.pop("timeout", 300)
        timeout = Timeout(
            connect=timeout_seconds,
            read=timeout_seconds,
        )
        # Default retries of 5 with exponential backoff
        retry = Retry(
            total=kwargs.pop("retries", 5),
            backoff_factor=kwargs.pop("backoff_factor", 0.25),
            status_forcelist=[500, 502, 503, 504],
        )
        # Configure certificate requirements for SSL connections
        cert_reqs = "CERT_NONE"
        ca_certs = None
        if not disable_cert:
            import certifi
            cert_reqs = "CERT_REQUIRED"
            ca_certs = get_environ_variable("SSL_CERT_FILE") or certifi.where()
        # Finally, create and return the PoolManager instance
        return PoolManager(
            num_pools=num_pools,
            maxsize=maxsize,
            timeout=timeout,
            retries=retry,
            cert_reqs=cert_reqs,
            ca_certs=ca_certs,
            **kwargs
        )



import enum

from fred.rest.router.config import RouterConfig
from fred.rest.router.catalog.default._base import RouterBaseMixin
from fred.rest.router.catalog.default._example import RouterExampleMixin

from fred.rest.router.catalog.interface import RouterCatalogInterface


class RouterCatalog(RouterCatalogInterface, enum.Enum):
    BASE = RouterConfig.auto(prefix="")(apply=RouterBaseMixin)
    EXAMPLE = RouterConfig.auto(prefix="/example")(apply=RouterExampleMixin)

    def get_kwargs(self) -> dict:
        match self:
            case RouterCatalog.EXAMPLE:
                # Disable the backend for the example router
                return {"disregard_backend": True}
            case _:
                return {}



from typing import Literal, Optional, overload

import torch

from fred.settings import get_environ_variable


def get_device(
        device: Optional[str] = None,
        env_fallback: Optional[str] = None,
        default: Optional[str] = None,
    ) -> str:
    env_default = get_environ_variable(
        env_fallback or "FRD_MLOPS_DEVICE",
        default=None
    )
    # Determine the device to use for model inference
    return device or env_default or default or (
        "cuda" if torch.cuda.is_available() else
        "mps" if torch.backends.mps.is_available() else
        "cpu"
    )


@overload
def get_dtype(
        dtype: Optional[str] = None,
        as_str: Literal[True] = True,
        env_fallback: Optional[str] = None,
        default: Optional[str] = None,
    ) -> str:
    ...


@overload
def get_dtype(
        dtype: Optional[str] = None,
        as_str: Literal[False] = False,
        env_fallback: Optional[str] = None,
        default: Optional[str] = None,
    ) -> torch.dtype:
    ...


def get_dtype(
        dtype: Optional[str] = None,
        as_str: bool = False,
        env_fallback: Optional[str] = None,
        default: Optional[str] = None,
    ) -> str | torch.dtype:
    """
    Determine the data type to use for model inference.
    Args:
        dtype (Optional[str]): The desired data type as a string (e.g., "float16", "float32").
        as_str (bool): If True, return the data type as a string; otherwise, return the torch.dtype object.
        env_fallback (Optional[str]): The name of the environment variable to use as a fallback for the data type. If not provided, defaults to "FRD_MLOPS_DTYPE".
        default (Optional[str]): The default data type to use if neither `dtype` nor the environment variable is set.
    Returns:
        str | torch.dtype: The data type as a string or torch.dtype object.
    """
    # Early exit for string return
    if as_str:
        env_default = get_environ_variable(
            env_fallback or "FRD_MLOPS_DTYPE",
            default=None
        )
        return dtype or env_default or default or (
            "float16" if torch.cuda.is_available() else
            "float32"
        )
    # Return the actual torch.dtype object if not as_str
    return getattr(
        torch,
        get_dtype(dtype=dtype, as_str=True),
        torch.float32,
    )




class ComponentInterface:
    
    def __or__(self, other: 'ComponentInterface') -> 'ComponentInterface':
        # union: d1 | d2 | d3  (build a group)
        # allow Node | Group and Group | Node chaining
        from fred.edag.comp._group import Group
        from fred.edag.comp._node import Node

        match self, other:
            case Node(), Node():
                return Group(nodes=[self, other])
            case Node(), Group():
                return Group(nodes=[self, *other.nodes])
            case Group(), Node():
                return Group(nodes=[*self.nodes, other])
            case Group(), Group():
                return Group(nodes=[*self.nodes, *other.nodes])
            case _:
                raise TypeError("| expects Node or Group")



import posixpath
from dataclasses import dataclass

from fred.settings import logger_manager
from fred.worker.runner.model._runner_spec import RunnerSpec
from fred.worker.runner.plugins.interface import PluginInterface
from fred.settings import get_environ_variable

logger = logger_manager.get_logger(name=__name__)


@dataclass(frozen=True, slots=True)
class RunpodPlugin(PluginInterface):

    def _execute(
            self,
            spec: RunnerSpec,
            **kwargs
        ):
        """Execute the runner specification on RunPod's serverless platform.
        Args:
            spec (RunnerSpec): The runner specification to execute.
            **kwargs: Additional keyword arguments for RunPod configuration, such as:
                - runpod_key: API key for RunPod.
                - runpod_url: Base URL for RunPod API.
                - runpod_serverless_id: The ID of the serverless function to invoke.
                - runpod_serverless_endpoint: The endpoint to use ('run' or 'runsync').
                - headers: Additional headers to include in the request.
        """
        import requests

        RUNPOD_KEY = (
            kwargs.get("runpod_key")
            or get_environ_variable(name="RUNPOD_KEY", default="")
        )
        RUNPOD_URL = (
            kwargs.get("runpod_url")
            or get_environ_variable(name="RUNPOD_URL", default="https://api.runpod.ai/v2")
        )
        RUNPOD_SERVERLESS_ID = (
            kwargs.get("runpod_serverless_id")
            or get_environ_variable(name="RUNPOD_SERVERLESS_ID", default=None)
        )
        RUNPOD_SERVERLESS_ENDPOINT = (
            # We do expect the specific endpoint to come from kwargs in most cases
            kwargs.get("runpod_serverless_endpoint")
            or get_environ_variable(name="RUNPOD_SERVERLESS_ENDPOINT", default="run")
        )
        if not RUNPOD_KEY:
            raise ValueError("RUNPOD_KEY is required to execute the runner on RunPod.")
        if not RUNPOD_SERVERLESS_ID:
            raise ValueError("RUNPOD_SERVERLESS_ID is required to execute the runner on RunPod.")
        if RUNPOD_SERVERLESS_ENDPOINT not in ["run", "runsync"]:
            # Right now we only support these two endpoints... We might add more in the future.
            # https://docs.runpod.io/serverless/endpoints/send-requests 
            raise ValueError("RUNPOD_SERVERLESS_ENDPOINT must be either 'run' or 'runsync'.")
        target_url = posixpath.join(
            RUNPOD_URL,
            RUNPOD_SERVERLESS_ID,
            RUNPOD_SERVERLESS_ENDPOINT,
        )
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {RUNPOD_KEY}",
            **kwargs.get("headers", {}),
        }
        start_event = spec.as_event(drop_id=True)
        response = requests.post(
            url=target_url,
            json=start_event,
            headers=headers,
        )
        return {
            "target_url": target_url,
            "spec_created_at": spec.created_at,
            "status_code": response.status_code,
            "response": response.json(),
            "ok": response.ok,
        }



from fred.rest.router.interface import RouterInterfaceMixin
from fred.rest.router.endpoint import RouterEndpointAnnotation


class RouterBaseMixin(RouterInterfaceMixin):

    @RouterEndpointAnnotation.set(
        path="/",
        methods=["GET"],
        summary="Base endpoint",
        description="A simple base endpoint to check service availability."
    )
    def base(self, include_telemetry: bool = False, **kwargs) -> dict:
        if include_telemetry:
            from fred.utils.runtime import RuntimeProfilingSnapshot
            kwargs["telemetry"] = RuntimeProfilingSnapshot.auto().to_dict()
        return {
            "ok": True,
            **kwargs
        }

    @RouterEndpointAnnotation.set(
        path="/passthrough",
        methods=["POST"],
        summary="Passthrough POST endpoint",
        description="An endpoint that returns all received parameters.",
    )
    def passthrough(self, **kwargs) -> dict:
        return kwargs



import os
import json
from dataclasses import dataclass, field
from typing import Optional
def load_policy(filename: str, path: Optional[str] = None, **kwargs) -> dict:
    path = path or os.path.join(os.path.dirname(__file__), "templates")
    filepath = os.path.join(path, filename)
    # Read the policy file as a string
    with open(filepath, "r", encoding="utf-8") as file:
        policy_str = file.read()
    # Replace placeholders in the policy with actual values from kwargs
    for param_key, param_val in kwargs.items():
        param_ref = "${{param_key}}".replace("param_key", param_key)
        policy_str = policy_str.replace(param_ref, param_val)
    return json.loads(policy_str)


@dataclass(frozen=True, slots=True)
class MinioPolicyLoader:
    title: str
    filename: str
    requires: list[str] = field(default_factory=list)

    def load(self, path: Optional[str] = None, **kwargs) -> dict:
        if (missing := set(self.requires) - set(kwargs.keys())):
            raise ValueError(f"Missing required parameters for policy '{self.filename}': {missing}")
        return load_policy(filename=self.filename, path=path, **kwargs)



import uuid
from inspect import Signature, signature, Parameter
from dataclasses import dataclass, field, asdict
from typing import Callable, Optional

from fred.edag.comp.interface import ComponentInterface


@dataclass(frozen=True, slots=True)
class NodeFun:
    fname: str
    inner: Callable
    signature: Signature
    

    @classmethod
    def auto(cls, function: Callable, name: Optional[str] = None) -> "NodeFun":
        fname = name or getattr(function, "__name__", "undefined")
        return cls(
            fname=fname,
            inner=function,
            signature=signature(function),
        )

    def validate_parameter_compliance(self, *args, **kwargs) -> dict:
        # Determine if the signature accepts **kwargs
        var_kwargs = any(
            p.kind == Parameter.VAR_KEYWORD
            for p in self.signature.parameters.values()
        )
        # Validate keywords against signature if not accepting **kwargs
        clean_kwargs = {
            k: v
            for k, v in kwargs.items()
            if k in self.signature.parameters
        } if not var_kwargs else kwargs
        # Bind arguments to signature
        bound = self.signature.bind_partial(*args, **clean_kwargs)
        bound.apply_defaults()
        # Return the bound arguments as a dictionary
        return {
            "args": bound.args,
            "kwargs": bound.kwargs,
        }

    def __call__(self, *args, **kwargs):
        params = self.validate_parameter_compliance(*args, **kwargs)
        return self.inner(*params["args"], **params["kwargs"])
    
    def __name__(self):
        return self.fname

    def __hash__(self):
        return hash((
            self.inner,
            tuple(self.signature.parameters.items()),
        ))


@dataclass(frozen=True, slots=True)
class Node(ComponentInterface):
    name: str
    key: str  # Output key
    nfun: NodeFun
    # TODO: let's make the 'params' a frozenset (i.e., frozenparams) instead of a dict to ensure immutability
    params: dict = field(default_factory=dict)
    nid: str = field(default_factory=lambda: str(uuid.uuid4()))
    _inplace: bool = False
    _explode: bool = False  # Whether this node's output should be exploded when used as input to another node

    def __hash__(self):
        obj = asdict(self)
        obj["nfun"] = self.nfun.__hash__()
        obj["params"] = frozenset((obj.get("params") or {}).keys())  # only hash keys to avoid unhashable values
        return hash(frozenset(obj.items()))

    def clone(self, **kwargs) -> "Node":
        # Verify if 'inplace' is set via '_inplace' or 'inplace' keys; otherwise, keep current value
        for key in ("inplace", "_inplace"):
            value = kwargs.pop(key, None)
            if isinstance(value, bool):
                kwargs["_inplace"] = value
                break
        else:
            kwargs["_inplace"] = self._inplace
        # Verify if 'explode' is set via '_explode' or 'explode' keys; otherwise, keep current value
        for key in ("explode", "_explode"):
            value = kwargs.pop(key, None)
            if isinstance(value, bool):
                kwargs["_explode"] = value
                break
        else:
            kwargs["_explode"] = self._explode
        # Create a new Node with updated attributes
        return self.__class__(
            **{
                "name": self.name,
                "key": self.key,
                "nfun": self.nfun,
                "params": self.params,
                **kwargs,
            },
            nid=str(uuid.uuid4()),  # Must have a new ID
        )

    def wrap(self, function: Callable) -> "Node":
        fname = getattr(function, "__name__", "undef_wrapper_function")
        return self.clone(
            nfun=NodeFun.auto(
                name=fname,
                function=lambda *args, **kwargs: function(self.fun(*args, **kwargs))
            ),
        )

    @classmethod
    def auto(
            cls,
            function: Callable,
            inplace: bool = False,
            explode: bool = False,
            fname: Optional[str] = None,
            name: Optional[str] = None,
            key: Optional[str] = None,
            **params,
    ):
        name = name or getattr(function, "__name__", "undefined")
        return cls(
            name=name,
            key=key or name,
            nfun=NodeFun.auto(function=function, name=fname),
            _inplace=inplace,
            _explode=explode,
            params=params,
        )

    @property
    def fun(self) -> Callable:
        return self.nfun

    def inplace(self) -> "Node":
        return self.clone(_inplace=True)
    
    def explode(self) -> "Node":
        return self.clone(_explode=True)

    @property
    def E(self) -> "Node":
        # Shortcut to set explode=True
        return self.explode()

    def __invert__(self) -> "Node":
        # Unary ~ operator to set explode=True
        return self.explode()

    def with_output(self, key: str) -> "Node":
        return self.with_alias(alias=self.name, key=key, keep_key=False)

    def with_alias(self, alias: str, key: Optional[str] = None, keep_key: bool = False) -> "Node":
        return self.__class__(
            name=alias,
            key=key or (self.key if keep_key else alias),
            nfun=self.nfun,
            params=self.params,
            _inplace=self._inplace,
            _explode=self._explode,
        )

    def with_params(self, update_key: Optional[str] = None, **params) -> "Node":
        return self.__class__(
            name=self.name,
            nfun=self.nfun,
            key=update_key or self.key,
            params={
                **self.params,
                **params,
            },
            _inplace=self._inplace,
            _explode=self._explode,
        )

    def execute(self, **kwargs):
        params = {
            **self.params,
            **kwargs
        }
        if self._inplace:
            return self.fun(**params)
        from fred.future.impl import Future
        return Future(self.fun, **params)



import time
from dataclasses import dataclass
from typing import Optional

from fred.future import Future
from fred.future.result import FutureResult
from fred.dao.comp.catalog import FredQueue
from fred.worker.runner.status import RunnerStatus
from fred.worker.runner.signal import RunnerSignal
from fred.worker.runner.backend import RunnerBackend
from fred.worker.runner.model.catalog import RunnerModelCatalog
from fred.worker.runner.settings import (
    FRD_RUNNER_BACKEND,
    FRD_RUNNER_REQUEST_QUEUE,
    FRD_RUNNER_RESPONSE_QUEUE,
)
from fred.settings import logger_manager

logger = logger_manager.get_logger(name=__name__)


@dataclass(frozen=True, slots=True)
class RunnerClient:
    _runner_backend: RunnerBackend
    req_queue: FredQueue
    res_queue: FredQueue


    @classmethod
    def auto(
            cls,
            queue_slug: Optional[str] = None,
            service_name: Optional[str] = None,
            **kwargs
        ) -> "RunnerClient":
        queue_slug = queue_slug or kwargs.pop("queue_slug", None) or (
            logger.warning("Queue slug not specified; defaulting to: 'demo'")
            or "demo"
        )
        queue_name_request = FRD_RUNNER_REQUEST_QUEUE or f"req:{queue_slug}"
        queue_name_response = FRD_RUNNER_RESPONSE_QUEUE or f"res:{queue_slug}"
        runner_backend = RunnerBackend.auto(
            service_name=service_name or FRD_RUNNER_BACKEND,
            **kwargs
        )
        return cls(
            _runner_backend=runner_backend,
            req_queue=runner_backend.queue(name=queue_name_request),
            res_queue=runner_backend.queue(name=queue_name_response),
        )

    @property
    def PING(self):
        return self.signal(RunnerSignal.PING)

    @property
    def STOP(self):
        return self.signal(RunnerSignal.STOP)

    def signal(self, signal: str | RunnerSignal):
        signal = RunnerSignal[signal.upper()] if isinstance(signal, str) else signal
        return signal.send(self.req_queue)

    def runner_info(self, runner_id: str) -> tuple[str, RunnerStatus]:
        runner_status = self._runner_backend.keyval(
            key=RunnerStatus.get_key(runner_id=runner_id)
        )
        if not (out := runner_status.get()):
            logger.warning(f"No status found for runner_id: '{runner_id}'")
            return ("", RunnerStatus.UNDEFINED)
        return RunnerStatus.parse_value(value=out)

    def runner_status(self, runner_id: str) -> RunnerStatus:
        _, status = self.runner_info(runner_id=runner_id)
        return status

    def runner_queue(self, runner_id: str) -> str:
        queue_slug, _ = self.runner_info(runner_id=runner_id)
        return queue_slug

    def runners(self) -> dict[str, Optional[str]]:
        return {
            key: self._runner_backend.keyval(key=key).get()
            for key in self._runner_backend.keyval.keys(pattern="frd:runner:*")
        }

    def futures(self) -> dict[str, Optional[str]]:
        return {
            key: self._runner_backend.keyval(key=key).get()
            for key in self._runner_backend.keyval.keys(pattern="frd:future:*:status")
        }

    def send(
            self,
            item: dict,
            req_uuid_hash: bool = False,
            item_uuid_hash: bool = False,
    ) -> str:
        item_instance = RunnerModelCatalog.ITEM.value.uuid(payload=item, uuid_hash=item_uuid_hash)
        request = item_instance.as_request(
            use_hash=req_uuid_hash,
            request_id=item.get("request_id"),
        )
        request.dispatch(request_queue=self.req_queue)
        return request.request_id
    
    @staticmethod
    def fetch_status(request_id: str) -> Optional[str]:
        return FutureResult._get_status_key(future_id=request_id).get()

    def _pullsync(
            self,
            request_id: str,
            retry_sync: int = 10,
            retry_delay: float = 0.2,
            retry_backoff_rate: float = 0.1,
            **kwargs,
    ):
        from fred.future.utils import pull_future_result

        if not self._is_ready_for_pullsync(
                request_id=request_id,
                retry_sync=retry_sync,
                retry_delay=retry_delay,
                retry_backoff_rate=retry_backoff_rate,
                # Always fail in pullsync to raise an exception if not ready.
                # In addition, this will be caught by the Future to propagate the exception.
                fail=True,
        ):
            raise ValueError(f"The provided request_id '{request_id}' is not ready for pullsync.")

        return pull_future_result(
            future_id=request_id,
            retry_delay=retry_delay,
            retry_backoff_rate=retry_backoff_rate,
            **kwargs
        )

    def pullsync(
            self,
            request_id: str,
            retry_sync: int = 10,
            retry_delay: float = 0.2,
            retry_backoff_rate: float = 0.1,
            **kwargs,
    ) -> Future:
        logger.warning(
            "Using 'pullsync' is to be deprecated soon; use the 'Future.subscribe' "
            "method instead (by default this is used on fetch_result with 'use_pullsync=False')."
        )
        return Future(
            function=self._pullsync,
            request_id=request_id,
            retry_sync=retry_sync,
            retry_delay=retry_delay,
            retry_backoff_rate=retry_backoff_rate,
            **kwargs
        )

    def _is_ready_for_pullsync(
            self,
            request_id: str,
            retry_sync: int = 10,
            retry_delay: float = 0.2,
            retry_backoff_rate: float = 0.1,
            fail: bool = False,
    ) -> bool:
        if not self.fetch_status(request_id=request_id):
            logger.info(f"No status found for request_id '{request_id}'")
            retry_kwargs = {
                "request_id": request_id,
                "retry_sync": retry_sync - 1,
                "retry_delay": retry_delay * (1 + retry_backoff_rate),  # Exponential backoff
                "retry_backoff_rate": retry_backoff_rate,
                "fail": fail,
            }
            if retry_sync:
                time.sleep(retry_delay)
                return self._is_ready_for_pullsync(**retry_kwargs)
            elif fail:
                logger.error(f"Failed to fetch status for request_id '{request_id}' after retries.")
                raise ValueError(f"No status found for request_id '{request_id}'")
            else:
                return False
        return True

    def fetch_result(
            self,
            request_id: str,
            now: bool = False,
            timeout: Optional[float] = None,
            use_pullsync: bool = False,
            **kwargs,
    ) -> Optional[dict]:
        future = self.pullsync(request_id=request_id, **kwargs) \
            if use_pullsync else Future.subscribe(future_id=request_id, **kwargs)
        if now:
            return future.getwhatevernow()
        return future.wait_and_resolve(timeout=timeout)



import io
import base64
from typing import Optional

from PIL.Image import Image

from fred.settings import logger_manager

logger = logger_manager.get_logger(__name__)


def save_image_to_path(image: Image, path: str) -> None:
    """
    Save a PIL Image to the specified file path.

    Args:
        image (Image): The PIL Image to save.
        path (str): The file path where the image will be saved.
    """
    image.save(path)


def image_to_b64(image: Image, format: Optional[str] = None) -> str:
    """
    Convert a PIL Image to a base64-encoded string.
    Args:
        image (Image): The PIL Image to convert.
        format (Optional[str]): The format to use for encoding (e.g., "PNG", "JPEG"). Defaults to "PNG" if None.
    Returns:
        str: The base64-encoded image string.
    """
    # Convert the image to a base64 string
    buffer = io.BytesIO()
    image.save(buffer, format=format or "PNG")
    return base64.b64encode(buffer.getvalue()).decode("utf-8")


def get_image_from_b64(image_b64: str) -> Image:
    """Load an image from a base64-encoded string.
    Args:
        image_b64 (str): The base64-encoded image string.
    Returns:
        Image: The loaded PIL Image.
    """
    import PIL.Image
    # Decode the base64 string to bytes
    image_data = base64.b64decode(image_b64)
    # Open the image using PIL
    return PIL.Image.open(io.BytesIO(image_data))


def get_image_from_path(path: str) -> Image:
    """
    Load an image from the specified file path.
    Args:
        path (str): The file path to load the image from.
    Returns:
        Image: The loaded PIL Image.
    """
    import PIL.Image
    # Open the image using PIL
    return PIL.Image.open(path)


def display_image_from_b64(image_b64: str) -> None:
    """
    Display a base64-encoded image.

    Args:
        image_b64 (str): The base64-encoded image string.
    """
    image = get_image_from_b64(image_b64=image_b64)
    image.show()



from dataclasses import dataclass
from typing import Optional

from fastapi import FastAPI, Depends
from fastapi.middleware.cors import CORSMiddleware

from fred.settings import (
    get_environ_variable,
    logger_manager,
)
from fred.rest.settings import (
    FRD_RESTAPI_HOST,
    FRD_RESTAPI_PORT,
    FRD_RESTAPI_LOGLEVEL,
    FRD_RESTAPI_DISABLE_AUTH,
)
from fred.rest.config import ServerRouterCatalogConfig
from fred.rest.auth import verify_key

logger = logger_manager.get_logger(name=__name__)


@dataclass(frozen=True, slots=True)
class FredServer:
    app: FastAPI
    include_routers: list[str]
    exclude_routers: list[str]
    router_classname: Optional[str] = None
    router_classpath: Optional[str] = None

    @classmethod
    def auto(cls, **kwargs) -> "FredServer":
        # Include routers by checking on keyword argument or environment variable
        include_routers = kwargs.pop("include_routers", None) or get_environ_variable(
            "FRD_RESTAPI_INCLUDE_ROUTERS",
            default=""
        )
        if isinstance(include_routers, str):
            include_routers = [
                name.upper()
                for router in include_routers.split(";")
                if (name := router.strip())
            ]
        # Exclude routers by checking on keyword argument or environment variable
        exclude_routers = kwargs.pop("exclude_routers", None) or get_environ_variable(
            "FRD_RESTAPI_EXCLUDE_ROUTERS",
            default=""
        )
        if isinstance(exclude_routers, str):
            exclude_routers = [
                name.upper()
                for router in exclude_routers.split(";")
                if (name := router.strip())
            ]
        # Auth dependencies setup
        auth_dependency = [] if FRD_RESTAPI_DISABLE_AUTH else [
            Depends(verify_key),
        ]
        kwargs["dependencies"] = kwargs.get("dependencies", []) + auth_dependency
        # Create FastAPI app instance
        app_instance = FastAPI(**kwargs)
        app_instance.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],  # TODO: Parameterize via env.variables
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

        router_classname = kwargs.pop("router_classname", None)
        router_classpath = kwargs.pop("router_classpath", None)
        return cls(
            app=app_instance,
            include_routers=include_routers,
            exclude_routers=exclude_routers,
            router_classname=router_classname,
            router_classpath=router_classpath,
        )

    def __post_init__(self):
        logger.info("Attempting to register routers...")
        logger.info("Included routers candidates: %s", self.include_routers or "ALL")
        logger.info("Excluded routers candidates: %s", self.exclude_routers or "NONE")
        srcc = ServerRouterCatalogConfig.auto(
            classname=self.router_classname,
            classpath=self.router_classpath,
        )
        for router_spec in srcc.catalog:
            name = router_spec.name.upper()
            if self.include_routers and name not in self.include_routers:
                logger.info(f"Skipping router '{name}' as it's not in the include list.")
                continue
            if self.exclude_routers and name in self.exclude_routers:
                logger.info(f"Skipping router '{name}' as it's in the exclude list.")
                continue
            logger.info(f"Registering router '{name}'.")
            kwargs = router_spec.get_kwargs()
            router_instance = router_spec.value.reference.auto(**kwargs)
            router_configs = router_spec.value.config.get_configs()
            # Register the router in the FastAPI app
            self.app.include_router(router_instance.router, **router_configs) 

    def start(self, **kwargs):
        import uvicorn

        server_kwargs = {
            "host": kwargs.pop("host", FRD_RESTAPI_HOST),
            "port": int(kwargs.pop("port", FRD_RESTAPI_PORT)),
            "log_level": FRD_RESTAPI_LOGLEVEL,
            **kwargs,
        }

        uvicorn.run(self.app, **server_kwargs)



from dataclasses import dataclass
from typing import Any, Optional

from PIL.Image import Image


class ImageOutputBackendInterface:
    metadata: dict = {}
    client: Optional[Any] = None  # Placeholder for clients like MinIO, S3, etc.


@dataclass(frozen=True, slots=True)
class ImageOutputInterface(ImageOutputBackendInterface):
    """Interface for image output handling."""
    image: Image

    @classmethod
    def auto(cls, **kwargs) -> 'ImageOutputInterface':
        return cls(**kwargs)

    @classmethod
    def from_path(cls, path: str) -> "ImageOutputInterface":
        from fred.utils.imops import get_image_from_path

        image = get_image_from_path(path=path)
        return cls(image=image)

    def save(self, path: str, format: str = "PNG"):
        """Save the image to a file."""
        self.image.save(path, format=format)

    def show(self):
        """Display the image."""
        self.image.show()

    def out(self, **kwargs) -> str:
        raise NotImplementedError("Subclasses must implement out method.")



from fred.utils.imout.interface import ImageOutputInterface


class OutputString(ImageOutputInterface):
    """String output handler for images."""

    def out(self) -> str:
        from fred.utils.imops import image_to_b64

        return image_to_b64(self.image)



import os
from dataclasses import dataclass, field
from typing import Callable, Optional

from fastapi import Request

from fred.settings import logger_manager

logger = logger_manager.get_logger(name=__name__)


@dataclass(frozen=True, slots=True)
class RouterEndpointConfig:
    path: Optional[str] = None
    methods: list[str] = field(default_factory=lambda: ["GET"])
    configs: dict = field(default_factory=dict)

    @classmethod
    def auto(cls, **kwargs) -> "RouterEndpointConfig":
        return cls(
            path=kwargs.pop("path", None),
            methods=(
                kwargs.pop("methods", None)
                or ["GET"]
            ),
            configs=kwargs
        ) 


@dataclass(frozen=True, slots=False)
class RouterEndpoint:
    function: Callable
    configs: RouterEndpointConfig
    _og: Optional[Callable] = None  # Original function before any binding... this is mainly for testing purposes.

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)

    def get_path(self, overwrite: Optional[str] = None) -> str:
        fname = self.function.__name__
        return (
            overwrite
            or self.configs.path
            or logger.debug(f"Missing 'path' definition, defaulting to function name: {fname}")
            or os.path.join("/", fname)
        )

    def route_config(self, path: Optional[str] = None) -> dict:
        return {
            "path": self.get_path(overwrite=path),
            "methods": self.configs.methods,
            **self.configs.configs,
        }

# Here are a couple of reference to understand why this class is needed:
# https://stackoverflow.com/questions/2365701/decorating-class-methods-how-to-pass-the-instance-to-the-decorator/3296318#3296318
# https://stackoverflow.com/questions/75409483/why-do-i-get-missing-required-argument-self-when-using-a-decorator-written-as
@dataclass(frozen=True, slots=True)
class RouterEndpointAnnotation(RouterEndpoint):
    
    @classmethod
    def set(cls, **kwargs) -> Callable:
        config = RouterEndpointConfig.auto(**kwargs)
        return lambda function: cls(function=function, configs=config)

    def __get__(self, other_instance, other_class) -> RouterEndpoint:
        # Create a partial function that binds the instance (other) to the method.
        # This method was originally implemented using the following approaches:
        # - function = functools.partial(self.function, other)
        # - function = lambda payload: self.function(other, **payload)
        # - function = functools.wraps(self.function)(lambda *args, **kwargs: self.function(other, *args, **kwargs)) 
        # However, these approaches did not work well with FastAPI's async handling and request parsing
        # since the dynamic lambda attributes were not being recognized properly by FastAPI.
        # The current implementation defines an async closure that correctly calls the inner function
        # and retrieves the parameters directly from the Request object.
        # For more info on how to use the request object, see:
        # - https://fastapi.tiangolo.com/advanced/using-request-directly/#use-the-request-object-directly
        # - https://stackoverflow.com/questions/67636088/how-to-access-request-object-in-router-function-using-fastapi
        # - https://www.starlette.dev/requests/
        async def closure(request: Request):
            params = {
                **request.headers,
                **request.path_params,
                **request.query_params,
            }
            try:
                if request.method in ("POST", "PUT", "PATCH"):
                    params.update(await request.json())
            except Exception:
                pass
            # Since the annotation is usually applied to methods within a class,
            # in most cases the 'other_instance' will be 'None'; therefore using 'other_class'
            # should allow accessing the shared-class level state (e.g., runner_backend).
            return self.function(other_class, **params)
        return RouterEndpoint(
            function=closure,
            configs=self.configs,
            _og=self.function,
        )
